==> camitk_3.4.0.orig.tar.gz: camitk-3.4.0-Source/sdk/actions/image/reconstruction/ImageReconstructionAction.cpp <==
/*****************************************************************************
 * $CAMITK_LICENCE_BEGIN$
 *
 * CamiTK - Computer Assisted Medical Intervention ToolKit
 * (c) 2001-2014 UJF-Grenoble 1, CNRS, TIMC-IMAG UMR 5525 (GMCAO)
 *
 * Visit http://camitk.imag.fr for more information
 *
 * This file is part of CamiTK.
 *
 * CamiTK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * CamiTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with CamiTK.  If not, see <http://www.gnu.org/licenses/>.
 *
 * $CAMITK_LICENCE_END$
 ****************************************************************************/

// -- Core image component stuff classes
#include "ImageReconstructionAction.h"
#include "ImageComponent.h"
#include "MeshComponent.h"

// -- Core stuff classes
#include "Application.h"
#include <Property.h>

// -- QT stuff
#include <QLineEdit>
#include <QFileDialog>
#include <QFileInfo>
#include <QMessageBox>

// -- stl stuff
#include <sstream>

// -- vtk filters stuff
#include <vtkImageResample.h>
#include <vtkContourFilter.h>
#include <vtkPolyDataNormals.h>
#include <vtkPolyDataConnectivityFilter.h>
#include <vtkPolyDataWriter.h>
#include <vtkMarchingCubes.h>
#include <vtkPointSet.h>
#include <vtkCallbackCommand.h>

#include <ImageComponent.h>
using namespace camitk;


// -------------------- constructor --------------------
ImageReconstructionAction::ImageReconstructionAction(ActionExtension* extension) : Action(extension) {
    this->setName("Reconstruction");
    this->setDescription("<br /> Marching cubes is a simple algorithm for creating a triangle mesh from an implicit function (one of the form f(x, y, z) = 0).<br /> It works by iterating (\"marching\") over a uniform grid of cubes superimposed over a region of the function. If all 8 vertices of the cube are positive, or all 8 vertices are negative, the cube is entirely above or entirely below the surface and no triangles are emitted. Otherwise, the cube straddles the function and some triangles and vertices are generated. Since each vertex can either be positive or negative, there are technically 28 possible configurations, but many of these are equivalent to one another. <br/> <br/>This action uses this algorithm to build a 3D surfacic mesh of the input image.");
    this->setComponent("ImageComponent");
    this->setFamily("Reconstruction");
    this->addTag("Reconstruction");

    // Action parameters
    Property* thresholdProperty = new Property(tr("Threshold"), 127.5, tr("Grey level value corresponding to the isovalue the isosurface represents."), "");
    thresholdProperty->setAttribute("minimum", 0.0);
    thresholdProperty->setAttribute("singleStep", 0.1);
    addParameter(thresholdProperty);

    // Do I keep the largest component ?
    Property* keepLargestCCProperty = new Property(tr("Keep only largest component?"), true, tr("Do we keep only the largest component?"), "");
    addParameter(keepLargestCCProperty);

    // Subsample original image ?
    Property* subsampleOrginalImageProperty = new Property(tr("Subsample original image?"), true, tr("Do we subsample input image for faster computation?"), "");
    addParameter(subsampleOrginalImageProperty);

    Property* subsampledImageX = new Property(tr("Subsampled image width (X)"), 64, tr("The width (X dimension) of the subsampled image"), "");
    subsampledImageX->setAttribute("minimum", 1);
    subsampledImageX->setAttribute("singleStep", 1);
    addParameter(subsampledImageX);

    Property* subsampledImageY = new Property(tr("Subsampled image height (Y)"), 64, tr("The height (Y dimension) of the subsampled image"), "");
    subsampledImageY->setAttribute("minimum", 1);
    subsampledImageY->setAttribute("singleStep", 1);
    addParameter(subsampledImageY);

    Property* subsampledImageZ = new Property(tr("Subsampled image depth (Z)"), 64, tr("The depth (Z dimension) of the subsampled image"), "");
    subsampledImageZ->setAttribute("minimum", 1);
    subsampledImageZ->setAttribute("singleStep", 1);
    addParameter(subsampledImageZ);
}

// -------------------- destructor --------------------
ImageReconstructionAction::~ImageReconstructionAction() {
    if (actionWidget)
        delete actionWidget;
}

// -------------------- apply --------------------
Action::ApplyStatus ImageReconstructionAction::apply() {

    // get original input ImageComponent
    this->myComponent = dynamic_cast<ImageComponent *>(getTargets().last());

    // compute model
    vtkSmartPointer<vtkPointSet> resultPointSet = this->getMarchingCubesReconstruction();
    new MeshComponent(resultPointSet, myComponent->getName() + "_mesh");
    Application::refresh();

    return SUCCESS;
}

// -------------------- getMarchingCubesReconstruction --------------------
vtkSmartPointer<vtkPointSet> ImageReconstructionAction::getMarchingCubesReconstruction() {
    double isoValue = property("Threshold").toDouble();
    bool keepLargestConnectedComponent = property("Keep only largest component?").toBool();
    bool subsample = property("Subsample original image?").toBool();
    int subSampledDimX = property("Subsampled image width (X)").toInt();
    int subSampledDimY = property("Subsampled image height (Y)").toInt();
    int subSampledDimZ = property("Subsampled image depth (Z)").toInt();

    Application::showStatusBarMessage("Computing Marching Cube Reconstruction");
    Application::resetProgressBar();
    vtkSmartPointer<vtkCallbackCommand> progressCallback = vtkSmartPointer<vtkCallbackCommand>::New();
    progressCallback->SetCallback(&Application::vtkProgressFunction);

    vtkSmartPointer<vtkImageData> originalImageData = this->myComponent->getImageData();
    vtkSmartPointer<vtkImageData> currentData = originalImageData;

    // For medical volumes, marching cubes generates a large number of triangles.
    // To be practical, we'll use a reduced resolution dataset.
    // For example, we take original 256^3 data and reduce it to 64^2 slices
    // by averaging neighboring pixels twice in the slice plane.
    // We call the resulting dataset quarter since it has 1/4 the resolution of the original data.
    vtkSmartPointer<vtkImageResample> resampleFilter = NULL;

    if (subsample) {
        resampleFilter = vtkSmartPointer<vtkImageResample>::New();
        int * previousDimensions = originalImageData->GetDimensions();

        double x_magnification = 0.0;
        double y_magnification = 0.0;
        double z_magnification = 0.0;
        x_magnification = subSampledDimX / (double) previousDimensions[0];
        y_magnification = subSampledDimY / (double) previousDimensions[1];

        resampleFilter->SetAxisMagnificationFactor(0, x_magnification);
        resampleFilter->SetAxisMagnificationFactor(1, y_magnification);

        if (originalImageData->GetDataDimension() == 3) {
            z_magnification = subSampledDimZ / (double) previousDimensions[2];
            resampleFilter->SetAxisMagnificationFactor(2, z_magnification);
        }

        resampleFilter->SetInput(originalImageData);
        resampleFilter->AddObserver(vtkCommand::ProgressEvent, progressCallback);

        currentData = resampleFilter->GetOutput();

        currentData->Update();
    }

    // The filter we have chosen to use to compute 3D isosurface of the volume image
    // is vtkMarchingcubes. We could also use vtkContourFilter since it will automatically
    // create an instance of vtkMarchingCubes as it delegates to the fastest subclass
    // for a particular dataset type.
    vtkSmartPointer<vtkMarchingCubes> mcSurface = vtkSmartPointer<vtkMarchingCubes>::New();
    mcSurface->SetInput(currentData);
    mcSurface->AddObserver(vtkCommand::ProgressEvent, progressCallback);

    mcSurface->SetValue(0, isoValue);

    // The class vtkPolyDataNormals is used to generate
    // nice surface normals for the data. vtkMarchingCubes can also generate normals,
    // but sometimes better results are achieved when the normals are directly from the
    // surface (vtkPolyDataNormals) versus from the data (vtkMarchingCubes).
    vtkSmartPointer<vtkPolyDataNormals> normalsFilter = vtkSmartPointer<vtkPolyDataNormals>::New();

    //Specify the angle that defines a sharp edge. If the difference in angle
    // across neighboring polygons is greater than this value,
    // the shared edge is considered "sharp".
    normalsFilter->SetFeatureAngle(60.0);

    // Keep largest connected component
    vtkSmartPointer<vtkPolyDataConnectivityFilter> connectivityFilter = NULL;

    if (keepLargestConnectedComponent) {
        connectivityFilter = vtkPolyDataConnectivityFilter::New();
        connectivityFilter->SetInput(mcSurface->GetOutput());
        connectivityFilter->SetExtractionModeToLargestRegion();
        normalsFilter->SetInput(connectivityFilter->GetOutput());
    } else
        normalsFilter->SetInput(mcSurface->GetOutput());

    normalsFilter->AddObserver(vtkCommand::ProgressEvent, progressCallback);

    vtkSmartPointer<vtkPointSet> resultPointSet = normalsFilter->GetOutput();
    resultPointSet->Update();

    // One can make efficient data visualization using vtkTriangleFilter
    // and vtkStripper. But first, let us try without...

    // Cleaning...
    originalImageData = NULL;
    currentData = NULL;

    Application::showStatusBarMessage("");
    Application::resetProgressBar();

    return resultPointSet;
}

==> camitk_4.0.4.orig.tar.gz: camitk-4.0.4-Source/sdk/actions/image/reconstruction/ImageReconstructionAction.cpp <==
/*****************************************************************************
 * $CAMITK_LICENCE_BEGIN$
 *
 * CamiTK - Computer Assisted Medical Intervention ToolKit
 * (c) 2001-2016 Univ. Grenoble Alpes, CNRS, TIMC-IMAG UMR 5525 (GMCAO)
 *
 * Visit http://camitk.imag.fr for more information
 *
 * This file is part of CamiTK.
 *
 * CamiTK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * CamiTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with CamiTK.  If not, see <http://www.gnu.org/licenses/>.
 *
 * $CAMITK_LICENCE_END$
 ****************************************************************************/

// -- Core image component stuff classes
#include "ImageReconstructionAction.h"
#include "ImageComponent.h"
#include "MeshComponent.h"

// -- Core stuff classes
#include "Application.h"
#include <Property.h>

// -- QT stuff
#include <QLineEdit>
#include <QFileDialog>
#include <QFileInfo>
#include <QMessageBox>

// -- stl stuff
#include <sstream>

// -- vtk filters stuff
#include <vtkImageResample.h>
#include <vtkContourFilter.h>
#include <vtkPolyDataNormals.h>
#include <vtkPolyDataConnectivityFilter.h>
#include <vtkPolyDataWriter.h>
#include <vtkMarchingCubes.h>
#include <vtkPointSet.h>
#include <vtkCallbackCommand.h>

#include <ImageComponent.h>
using namespace camitk;


// -------------------- constructor --------------------
ImageReconstructionAction::ImageReconstructionAction(ActionExtension* extension) : Action(extension) {
    this->setName("Reconstruction");
    this->setDescription("<br /> Marching cubes is a simple algorithm for creating a triangle mesh from an implicit function (one of the form f(x, y, z) = 0).<br /> It works by iterating (\"marching\") over a uniform grid of cubes superimposed over a region of the function. If all 8 vertices of the cube are positive, or all 8 vertices are negative, the cube is entirely above or entirely below the surface and no triangles are emitted. Otherwise, the cube straddles the function and some triangles and vertices are generated. Since each vertex can either be positive or negative, there are technically 28 possible configurations, but many of these are equivalent to one another. <br/> <br/>This action uses this algorithm to build a 3D surfacic mesh of the input image.");
    this->setComponent("ImageComponent");
    this->setFamily("Reconstruction");
    this->addTag("Reconstruction");

    // Action parameters
    Property* thresholdProperty = new Property(tr("Threshold"), 127.5, tr("Grey level value corresponding to the isovalue the isosurface represents."), "");
    thresholdProperty->setAttribute("minimum", 0.0);
    thresholdProperty->setAttribute("singleStep", 0.1);
    addParameter(thresholdProperty);

    // Do I keep the largest component ?
    Property* keepLargestCCProperty = new Property(tr("Keep only largest component?"), true, tr("Do we keep only the largest component?"), "");
    addParameter(keepLargestCCProperty);

    // Subsample original image ?
    Property* subsampleOrginalImageProperty = new Property(tr("Subsample original image?"), true, tr("Do we subsample input image for faster computation?"), "");
    addParameter(subsampleOrginalImageProperty);

    Property* subsampledImageX = new Property(tr("Subsampled image width (X)"), 64, tr("The width (X dimension) of the subsampled image"), "");
    subsampledImageX->setAttribute("minimum", 1);
    subsampledImageX->setAttribute("singleStep", 1);
    addParameter(subsampledImageX);

    Property* subsampledImageY = new Property(tr("Subsampled image height (Y)"), 64, tr("The height (Y dimension) of the subsampled image"), "");
    subsampledImageY->setAttribute("minimum", 1);
    subsampledImageY->setAttribute("singleStep", 1);
    addParameter(subsampledImageY);

    Property* subsampledImageZ = new Property(tr("Subsampled image depth (Z)"), 64, tr("The depth (Z dimension) of the subsampled image"), "");
    subsampledImageZ->setAttribute("minimum", 1);
    subsampledImageZ->setAttribute("singleStep", 1);
    addParameter(subsampledImageZ);
}

// -------------------- destructor --------------------
ImageReconstructionAction::~ImageReconstructionAction() {
    if (actionWidget) {
        delete actionWidget;
    }
}

// -------------------- apply --------------------
Action::ApplyStatus ImageReconstructionAction::apply() {

    // get original input ImageComponent
    this->myComponent = dynamic_cast<ImageComponent*>(getTargets().last());

    // compute model
    vtkSmartPointer<vtkPointSet> resultPointSet = this->getMarchingCubesReconstruction();

    MeshComponent* outputComp = new MeshComponent(resultPointSet, myComponent->getName() + "_mesh");

    // consider frame policy on new image created
    Action::applyTargetPosition(myComponent, outputComp);

    Application::refresh();

    return SUCCESS;
}

// -------------------- getMarchingCubesReconstruction --------------------
vtkSmartPointer<vtkPointSet> ImageReconstructionAction::getMarchingCubesReconstruction() {
    double isoValue = property("Threshold").toDouble();
    bool keepLargestConnectedComponent = property("Keep only largest component?").toBool();
    bool subsample = property("Subsample original image?").toBool();
    int subSampledDimX = property("Subsampled image width (X)").toInt();
    int subSampledDimY = property("Subsampled image height (Y)").toInt();
    int subSampledDimZ = property("Subsampled image depth (Z)").toInt();

    Application::showStatusBarMessage("Computing Marching Cube Reconstruction");
    Application::resetProgressBar();
    vtkSmartPointer<vtkCallbackCommand> progressCallback = vtkSmartPointer<vtkCallbackCommand>::New();
    progressCallback->SetCallback(&Application::vtkProgressFunction);

    // The filter we have chosen to use to compute 3D isosurface of the volume image
    // is vtkMarchingcubes. We could also use vtkContourFilter since it will automatically
    // create an instance of vtkMarchingCubes as it delegates to the fastest subclass
    // for a particular dataset type.
    vtkSmartPointer<vtkMarchingCubes> mcSurface = vtkSmartPointer<vtkMarchingCubes>::New();
 
    // For medical volumes, marching cubes generates a large number of triangles.
    // To be practical, we'll use a reduced resolution dataset.
    // For example, we take original 256^3 data and reduce it to 64^2 slices
    // by averaging neighboring pixels twice in the slice plane.
    // We call the resulting dataset quarter since it has 1/4 the resolution of the original data.
    vtkSmartPointer<vtkImageResample> resampleFilter = NULL;

    if (subsample) {
        resampleFilter = vtkSmartPointer<vtkImageResample>::New();
        int* previousDimensions = myComponent->getImageData()->GetDimensions();

        double x_magnification = 0.0;
        double y_magnification = 0.0;
        double z_magnification = 0.0;
        x_magnification = subSampledDimX / (double) previousDimensions[0];
        y_magnification = subSampledDimY / (double) previousDimensions[1];

        resampleFilter->SetAxisMagnificationFactor(0, x_magnification);
        resampleFilter->SetAxisMagnificationFactor(1, y_magnification);

        if (myComponent->getImageData()->GetDataDimension() == 3) {
            z_magnification = subSampledDimZ / (double) previousDimensions[2];
            resampleFilter->SetAxisMagnificationFactor(2, z_magnification);
        }

        // setup the initial data for the pipeline
        resampleFilter->SetInputData(myComponent->getImageData());
        resampleFilter->AddObserver(vtkCommand::ProgressEvent, progressCallback);
   
        // setup next step in the pipeline
        mcSurface->SetInputConnection(resampleFilter->GetOutputPort());
    }
    else {
        // setup the initial data for the pipeline
        mcSurface->SetInputData(myComponent->getImageData());
    }
    
    mcSurface->AddObserver(vtkCommand::ProgressEvent, progressCallback);
    mcSurface->SetValue(0, isoValue);

    // The class vtkPolyDataNormals is used to generate
    // nice surface normals for the data. vtkMarchingCubes can also generate normals,
    // but sometimes better results are achieved when the normals are directly from the
    // surface (vtkPolyDataNormals) versus from the data (vtkMarchingCubes).
    vtkSmartPointer<vtkPolyDataNormals> normalsFilter = vtkSmartPointer<vtkPolyDataNormals>::New();

    //Specify the angle that defines a sharp edge. If the difference in angle
    // across neighboring polygons is greater than this value,
    // the shared edge is considered "sharp".
    normalsFilter->SetFeatureAngle(60.0);

    // Keep largest connected component
    vtkSmartPointer<vtkPolyDataConnectivityFilter> connectivityFilter = NULL;

    if (keepLargestConnectedComponent) {
        connectivityFilter = vtkPolyDataConnectivityFilter::New();
        connectivityFilter->SetInputConnection(mcSurface->GetOutputPort());
        connectivityFilter->SetExtractionModeToLargestRegion();
        // setup next step in the pipeline
        normalsFilter->SetInputConnection(connectivityFilter->GetOutputPort());
    }
    else {
        // setup next step in the pipeline
        normalsFilter->SetInputConnection(mcSurface->GetOutputPort());
    }

    normalsFilter->AddObserver(vtkCommand::ProgressEvent, progressCallback);
    normalsFilter->Update();
    
    vtkSmartPointer<vtkPointSet> resultPointSet = normalsFilter->GetOutput();

    // One can make efficient data visualization using vtkTriangleFilter
    // and vtkStripper. But first, let us try without...

    Application::showStatusBarMessage("");
    Application::resetProgressBar();

    return resultPointSet;
}
